package de.intevation.flys.exports;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;

import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Line2D;

import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.Range;
import org.jfree.data.xy.XYSeries;
import org.jfree.ui.TextAnchor;
import org.jfree.chart.plot.PlotRenderingInfo;
import org.jfree.text.TextUtilities;
import org.jfree.chart.entity.XYAnnotationEntity;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.ui.RectangleEdge;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.util.LineUtilities;

import org.w3c.dom.Document;

import de.intevation.artifacts.Artifact;

import de.intevation.artifactdatabase.state.Facet;

import de.intevation.flys.artifacts.FLYSArtifact;

import de.intevation.flys.artifacts.model.FacetTypes;
import de.intevation.flys.artifacts.model.WQKms;

import de.intevation.flys.model.Annotation;

/**
 * Custom Annotations class that is drawn only if no collisions with other
 * already drawn CustomAnnotations in current plot are found.
 * Draws a given text and a line to it from either axis.
 */
class CustomAnnotation extends XYTextAnnotation {

    /** Logger for this class. */    
    private static Logger logger =
        Logger.getLogger(CustomAnnotation.class);

    /** Simplified view on axes. */
    public static enum SimpleAxis {
        X_AXIS, /** Usually "horizontal". */
        Y_AXIS  /** Usually "vertical". */
    }

    /** Which axis to stick to. */
    protected SimpleAxis stickyAxis = SimpleAxis.X_AXIS;


    /**
     * Trivial constructor.
     *
     * @param text Text to display.
     * @param x    X-position in dataspace (typical horizontal, in km).
     * @param y    Y-position in dataspace (typical vertical, in m).
     */
    public CustomAnnotation(String text, float x, float y) {
        super(text, x, y);
    }


    /**
     * Sets the "sticky axis" (whether to draw annotations at the
     * X- or the Y-Axis.
     *
     * @param stickyAxis axis to stick to.
     */
    public void setStickyAxis(SimpleAxis stickyAxis) {
        this.stickyAxis = stickyAxis;
    }


    /**
     * Draws a small line at axis where this annotation resides.
     *
     * @param g2          the graphics device.
     * @param dataArea    the data area.
     * @param domainAxis  the domain axis.
     * @param rangeAxis   the range axis.
     * @param domainEdge  the domain edge.
     * @param rangeEdge   the range edge.
     * @param orientation the plot orientation.
     */
    protected void drawAxisMark(
            java.awt.Graphics2D g2,
            java.awt.geom.Rectangle2D dataArea,
            ValueAxis domainAxis,
            ValueAxis rangeAxis,
            RectangleEdge domainEdge,
            RectangleEdge rangeEdge,
            PlotOrientation orientation) {
        float j2DX1 = 0.0f;
        float j2DX2 = 0.0f;
        float j2DY1 = 0.0f;
        float j2DY2 = 0.0f;
        float x = (float) getX();
        float y = (float) getY();
        /* When dependent on X/Y-Axis and orientation, following
           can be used as a base:
        if (orientation == PlotOrientation.VERTICAL) {
            j2DX1 = (float) domainAxis.valueToJava2D(x, dataArea,
                        domainEdge);
            j2DY1 = (float) rangeAxis.valueToJava2D(y, dataArea,
                        rangeEdge);
            j2DX2 = (float) domainAxis.valueToJava2D(x, dataArea,
                        domainEdge);
            j2DY2 = (float) rangeAxis.valueToJava2D(y, dataArea,
                        rangeEdge);
            }
        else if (orientation == PlotOrientation.HORIZONTAL) {
            j2DY1 = (float) domainAxis.valueToJava2D(x, dataArea,
                    domainEdge);
            j2DX1 = (float) rangeAxis.valueToJava2D(y, dataArea,
                    rangeEdge);
            j2DY2 = (float) domainAxis.valueToJava2D(x, dataArea,
                    domainEdge);
            j2DX2 = (float) rangeAxis.valueToJava2D(y, dataArea,
                    rangeEdge);
        }

        g2.setPaint(this.paint);
        g2.setStroke(this.stroke);
        */
        j2DY1 = (float) RectangleEdge.coordinate(dataArea, domainEdge);
        j2DY2 = j2DY1 - 0.10f * (float)
            (rangeAxis.getRange().getUpperBound() 
             - rangeAxis.getRange().getLowerBound());
        j2DX1 = (float) domainAxis.valueToJava2D(x, dataArea, domainEdge);
        j2DX2 = j2DX1;

        Line2D line = new Line2D.Float(j2DX1, j2DY1, j2DX2, j2DY2);

        // line is clipped to avoid JRE bug 6574155, for more info
        // see JFreeChart bug 2221495
        boolean visible = LineUtilities.clipLine(line, dataArea);
        if (visible) {
            g2.draw(line);
        }
    }


    /**
     * Draw the Annotiation if it does not collide with other already drawn
     * Annotations.
     *
     * @param g2            the graphics device.
     * @param plot          the plot.
     * @param dataArea      the data area.
     * @param domainAxis    the domain axis.
     * @param rangeAxis     the range axis.
     * @param rendererIndex the render index.
     * @param info          state information, escpecially collects info about
     *                      already drawn shapes (and thus annotations), used
     *                      for collision detection.
     */
    @Override
    public void draw(
       java.awt.Graphics2D g2,
       XYPlot plot,
       java.awt.geom.Rectangle2D dataArea,
       ValueAxis domainAxis,
       ValueAxis rangeAxis,
       int rendererIndex,
       PlotRenderingInfo info) {

        if (info == null)
            return;

        // Calculate the bounding box.
        ChartRenderingInfo chartInfo = info.getOwner();

        PlotOrientation orientation = plot.getOrientation();
        RectangleEdge domainEdge = Plot.resolveDomainAxisLocation(
            plot.getDomainAxisLocation(), orientation);
        RectangleEdge rangeEdge = Plot.resolveRangeAxisLocation(
            plot.getRangeAxisLocation(), orientation);
        float anchorX = (float) domainAxis.valueToJava2D(
            getX(), dataArea, domainEdge);
        float anchorY = (float) rangeAxis.valueToJava2D(
            getY(), dataArea, rangeEdge);
        if (orientation == PlotOrientation.HORIZONTAL) {
            float tempAnchor = anchorX;
            anchorX = anchorY;
            anchorY = tempAnchor;
        }

        // Always draw the small line at axis.
        drawAxisMark(g2, dataArea, domainAxis, rangeAxis, domainEdge,
                rangeEdge, orientation);

        g2.setFont(getFont());
        Shape hotspot = TextUtilities.calculateRotatedStringBounds(
           getText(), g2, anchorX, anchorY, getTextAnchor(),
           getRotationAngle(), getRotationAnchor());
        Rectangle2D hotspotBox = hotspot.getBounds2D();
        // Check for collisions with other XYAnnotations.
        for (Iterator i = chartInfo.getEntityCollection().iterator();
                i.hasNext(); ) {
            Object next = i.next();
            // Collision with other stuff than XYAnnotations are okay.
            if (next instanceof XYAnnotationEntity) {
                XYAnnotationEntity drawnShape = (XYAnnotationEntity) next;
                if (drawnShape.getArea().intersects(hotspotBox)) {
                    // Found collision, early stop.
                    return;
                }
            }
        }

        // Actuall drawing.
        if (getBackgroundPaint() != null) {
        g2.setPaint(getBackgroundPaint());
            g2.fill(hotspot);
        }
        g2.setPaint(getPaint());
        TextUtilities.drawRotatedString(getText(), g2, anchorX, anchorY,
                getTextAnchor(), getRotationAngle(), getRotationAnchor());
        // Draw outline.
        if (false) {
            g2.setStroke(getOutlineStroke());
            g2.setPaint(getOutlinePaint());
            g2.draw(hotspot);
        }

        // Add info that we have drawn this Annotation.
        addEntity(info, hotspot, rendererIndex, getToolTipText(), getURL());
    }
}


/**
 * An OutGenerator that generates discharge curves.
 *
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class LongitudinalSectionGenerator
extends      XYChartGenerator
implements   FacetTypes
{
    /** The logger that is used in this generator. */
    private static Logger logger =
        Logger.getLogger(LongitudinalSectionGenerator.class);

    public static final String I18N_CHART_TITLE =
        "chart.longitudinal.section.title";

    public static final String I18N_CHART_SUBTITLE =
        "chart.longitudinal.section.subtitle";

    public static final String I18N_XAXIS_LABEL =
        "chart.longitudinal.section.xaxis.label";

    public static final String I18N_YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.label";

    public static final String I18N_2YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.second.label";

    public static final String I18N_CHART_TITLE_DEFAULT  = "W-L\u00e4ngsschnitt";
    public static final String I18N_XAXIS_LABEL_DEFAULT  = "km";
    public static final String I18N_YAXIS_LABEL_DEFAULT  = "W [NN + m]";
    public static final String I18N_2YAXIS_LABEL_DEFAULT = "Q [m\u00b3/s]";


    protected boolean inverted;

    protected List<Annotation> annotations;

    public LongitudinalSectionGenerator() {
        super();
        annotations = new ArrayList<Annotation>();
    }


    protected String getChartTitle() {
        return msg(I18N_CHART_TITLE, I18N_CHART_TITLE_DEFAULT);
    }

    public boolean isInverted() {
        return inverted;
    }

    public void setInverted(boolean inverted) {
        this.inverted = inverted;
    }

    @Override
    protected void addSubtitles(JFreeChart chart) {
        double[] dist  = getRange();

        Object[] args = new Object[] {
            getRiverName(),
            dist[0],
            dist[1]
        };

        String subtitle = msg(I18N_CHART_SUBTITLE, "", args);
        chart.addSubtitle(new TextTitle(subtitle));
    }

    @Override
    public JFreeChart generateChart() {
        JFreeChart c = super.generateChart();
        XYPlot p = (XYPlot) c.getPlot();

        redoAnnotations(p, p.getDomainAxis());

        return c;
    }


    protected String getXAxisLabel() {
        return msg(I18N_XAXIS_LABEL, I18N_XAXIS_LABEL_DEFAULT);
    }


    protected String getYAxisLabel() {
        return msg(I18N_YAXIS_LABEL, I18N_YAXIS_LABEL_DEFAULT);
    }


    protected void adjustAxes(XYPlot plot) {
        super.adjustAxes(plot);

        NumberAxis qAxis = new NumberAxis(
            msg(I18N_2YAXIS_LABEL, I18N_2YAXIS_LABEL_DEFAULT));

        plot.setRangeAxis(1, qAxis);

        invertXAxis(plot.getDomainAxis());
    }


    /**
     * Remove all annotations from plot and re-insert them at an approximately
     * okay position. The followed approach is naive but side-effect free.
     *
     * @param plot the plot.
     * @param axis the value axis.
     */
    protected void redoAnnotations(XYPlot plot, ValueAxis axis) {
        plot.clearAnnotations();
        // TODO Position calculation could/should be done in
        // the CustomAnnotation-Implementation itself.
        ValueAxis yAxis = plot.getRangeAxis();
        float posY = 140.f;
        if (yAxis != null) {
            posY = (float) yAxis.getRange().getLowerBound();
            posYLess = posY;
            // Add some (2%) space between Text and axis.
            posY += 0.02f * (yAxis.getRange().getUpperBound()
                    - yAxis.getRange().getLowerBound());
        }

        // Add all annotations.
        for (Annotation a: annotations) {
            float posX = (float) a.getRange().getA().doubleValue();
            String text = a.getPosition().getValue();

            XYTextAnnotation ta = new CustomAnnotation(text, posX, posY);
            double rotation = 270.0f * (Math.PI / 180.0f);
            ta.setRotationAngle(rotation);
            ta.setRotationAnchor(TextAnchor.CENTER_LEFT);
            ta.setTextAnchor(TextAnchor.CENTER_LEFT);
            plot.getRenderer().addAnnotation(ta);
        }
    }


    /**
     * This method overrides the XYChartGenerators zoomY method to include the 0
     * value on the Q axis.
     */
    @Override
    protected boolean zoomY(XYPlot plot, ValueAxis axis, Range range, Range x) {
        if (plot.getRangeAxisIndex(axis) == 1) {
            // we want the Q axis to start at 0 if no zooming has been done
            range = new Range(0d, range.getUpperBound());
        }

        return super.zoomY(plot, axis, range, x);
    }


    /**
     * This method inverts the x-axis based on the kilometer information of the
     * selected river. If the head of the river is at kilometer 0, the axis is
     * not inverted, otherwise it is.
     *
     * @param xaxis The domain axis.
     */
    protected void invertXAxis(ValueAxis xaxis) {

        if (inverted) {
            logger.debug("Invert X-Axis.");
            xaxis.setInverted(true);
        }
    }


    public void doOut(Artifact artifact, Facet facet, Document attr) {
        String name = facet.getName();

        logger.debug("LongitudinalSectionGenerator.doOut: " + name);

        if (name == null) {
            logger.error("No facet name for doOut(). No output generated!");
            return;
        }

        FLYSArtifact flys = (FLYSArtifact) artifact;
        Facet        f    = flys.getNativeFacet(facet);

        if (f == null) {
            return;
        }

        if (name.equals(LONGITUDINAL_W)) {
            doWOut((WQKms) f.getData(artifact, context), attr);
        }
        else if (name.equals(LONGITUDINAL_Q)) {
            doQOut((WQKms) f.getData(artifact, context), attr);
        }
        else if (name.equals(LONGITUDINAL_ANNOTATION)) {
            doAnnotationsOut(f.getData(artifact, context), attr);
        }
        else {
            logger.warn("Unknown facet name: " + name);
            return;
        }
    }


    /**
     * Register annotations available for the diagram.
     *
     * @param o     list of annotations (data of facet).
     * @param theme ignored.
     */
    protected void doAnnotationsOut(Object o, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doAnnotationsOut");
        this.annotations = (List<Annotation>) o;
    }


    /**
     * Process the output for W facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     * @param theme The theme that contains styling information.
     */
    protected void doWOut(WQKms wqkms, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doWOut");

        XYSeries series = new StyledXYSeries(getSeriesName(wqkms, "W"), theme);

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKm(0));
                logger.debug("End   km: " + wqkms.getKm(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKm(i), wqkms.getW(i));
        }

        addFirstAxisSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    /**
     * Process the output for Q facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     * @param theme The theme that contains styling information.
     */
    protected void doQOut(WQKms wqkms, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doQOut");

        XYSeries series = new StyledXYSeries(getSeriesName(wqkms, "Q"), theme);

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKm(0));
                logger.debug("End   km: " + wqkms.getKm(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKm(i), wqkms.getQ(i));
        }

        addSecondAxisSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    protected String getSeriesName(WQKms wqkms, String mode) {
        String name   = wqkms.getName();
        String prefix = name != null && name.indexOf(mode) >= 0 ? null : mode;

        return prefix != null && prefix.length() > 0
            ? prefix + "(" + name +")"
            : name;
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
